/* arvore.c
 *
 * Linguagem C - rvores Binrias
 * Viva O Linux
 *
 * Enzo Ferber
 * 2015
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define FOLHA		(Folha)malloc(sizeof(struct folha))
#define CMDSTR		500

typedef struct folha* Folha;

struct folha {
	int 	info;

	Folha	esquerda;
	Folha	direita;
}; 

Folha inserir (Folha, int), deletar (Folha, int), procurar (Folha, int);
void imprimir_arvore (Folha, int);
void ordenada(Folha), preordenada(Folha), posordenada(Folha);
void cmd (void), help (void);

/* @ Folha inserir (Folha raiz, int info)
 *
 * Argumentos
 * ----------
 * 	raiz	ponetiro para 'struct folha', estrutura de dado bsica para
 * 		construo da rvore
 * 	info	nova informao a ser inserida
 *
 * Retorno
 * -------
 * 	NULL	em caso de erro
 * 	Folha	em caso de sucesso (n 'raiz')
 */
Folha inserir (Folha raiz, int info)
{
	if (!raiz) {
		/* significa que o ponteiro  nulo, e que est  a posio
		 * para insero.
		 *
		 * 1. Alocar e checar memria
		 */
		if (!(raiz = FOLHA)) {
			perror("inserir:malloc()");
			return NULL;
		}

		/* 2. Cpia da Informao
		 * 3. Ponteiros de referncia
		 * 4. Retorno da nova folha
		 */
		raiz->info = info;
		raiz->esquerda = raiz->direita = NULL;
		
		return raiz;
	} else if (info > raiz->info)
		raiz->direita = inserir (raiz->direita, info);
	else 
		raiz->esquerda = inserir (raiz->esquerda, info);

	/* retorna o ponteiro raiz
	 *
	 * Isso  necessrio pois a funo inserir  recursiva, e seu retorno
	 *  sempre atribuido ao mesmo ponteiro passado como argumento 'raiz',
	 *
	 * raiz->esquerda = inserir(raiz->esquerda,info)
	 * raiz->direita  = inserir(raiz->direta,  info)
	 */
	return raiz;
}

/* @ Folha deletar (Folha raiz, int info)
 *
 * Argumentos
 * ----------
 *  	raiz	raiz principal da arvore
 *  	info	informao procurada para deletar
 *
 *  Retorno
 *  -------
 *  	raiz	em ambos os casos
 */
Folha deletar (Folha raiz, int info) {
	/* para deletar um n,  necessrio que esse n exista.
	 * Ento, vamos procurar pelo n.
	 *
	 * A remoo de folhas em uma rvore  um pouco mais detalhada que sua
	 * similar para listas encadeadas. 
	 *
	 * Existem alguns pontos a se considerar para remover uma folha:
	 * 	1. A folha  tambm uma raiz (contm subrvores)?
	 * 	2. 
	 */
	Folha filho, n_raiz;

	/* se chegou ao final da arvore e nao encontrou nada para deletar....
	 */
	if (!raiz) return NULL;

	/* primeira comparao do programa:
	 * 	Achamos a informao procurada?
	 */
	if (raiz->info == info) {
		/* 1. Existe uma raiz direita?
		 * 	(uma raiz direita contm todos os elementos MAIORES que
		 * 	a raiz que estamos excluindo).
		 *
		 * 2. Caso ela exista, ela ser a nova raiz.
		 * 3. Os atuais elementos *menores* devero ser anexados nos
		 *    menores elementos da nova raiz (segundo passo).
		 */
		if (raiz->direita) {
			/* 2. a nova raiz  a subrvore da direita, que contem
			 * todos os elementos MAIORES que a raiz a ser removida
			 */
			n_raiz = filho = raiz->direita;
			
			/* 3. a subrvore esquerda da nova raz  percorrida 
			 * at onde no houverem mais elementos. Isso nos dar
			 * o menor elemento maior que a folha esquerda de nossa
			 * raiz a ser removida.
			 */ 
			while(filho->esquerda)
				filho = filho->esquerda;

			/* o menor elemento maior que a raiz
			 * torna-se
			 * raiz de todos os elementos menores que a raiz 
			 * removida
			 */
			filho->esquerda = raiz->esquerda;

			/* libera a memria do ponteiro */
			free (raiz);

			/* 1 + 2: nova raiz */
			return n_raiz;

		} else {
			/* caso no haja uma raiz direita (elementos MAIORES)
			 * que nossa raiz, a nova raiz da rvore seja o nico
			 * elemento restante: a subrvore esquerda da rvore,
			 * ou os antigos elementos menores que a raiz removida.
			 */
			n_raiz = raiz->esquerda;

			/* libera a memria do ponteiro */
			free (raiz);

			/* retorno da nova raiz */
			return n_raiz;
		}

	} else if (info > raiz->info) {
		/* Caso a informao procurada seja maior que a informao na
		 * raiz, devemos ento procurar pela informao na subrvore
		 * da direita (esta operao dever ser congruente em todo o 
		 * seu programa.
		 *
		 * Voc no pode ordenar a insero de um jeito e a remoo de
		 * outro!
		 */
		raiz->direita = deletar(raiz->direita, info);
	} else {
		/* Caso a informao seja menor que a raiz, procuraremos por ela
		 * nas subrvores da esquerda.
		 */
		raiz->esquerda = deletar(raiz->esquerda, info);
	}
	/* Lembre-se que as rvores binrias so estruturas de dados recursivas,
	 * e por esse motivo,  sempre bom saber o que se est retornando.
	 *
	 * Assim como na funo de insero, as chamadas recursivas  deletar()
	 * esperam como retorno uma raz! Para onde a raiz->esquerda vai 
	 * apontar quando a funo retornar?
	 *
	 * 	1. Se ela no for excluida, vai ser a prpria raiz->esquerda
	 * 	2. Caso seja excluda, ser determinada pelo algoritmo no bloco
	 * 	   if(raiz->info == info).
	 *
	 * Aqui, no final do escopo da funo, significa que  um retorno de
	 * chamadas recursivas  deletar (j que o bloco raiz->info == info)
	 * retorna por si s. J que este  um retorno das chamadas recursivas,
	 * deve retornar a raiz na qual foi chamada.
	 */
	return raiz;
}

Folha procurar (Folha raiz, int info)
{
	/* como todas as funes relativas a arvores binrias so recursivas,
	 * todas necessitam de um parametro de "parada de recurso", ento,
	 * sempre que acharmos um n NULO, retornamos (at porque no d pra
	 * imprimir informaes de um n que no existe.
	 */
	if (!raiz) return NULL;

	if (info > raiz->info) return procurar(raiz->direita, info);
	else return procurar (raiz->esquerda, info);

	/* falhas, compilao, etc */
	return NULL;
}

void imprimir_arvore (Folha raiz, int level)
{
	register int i;

	/* como todas as funes relativas a arvores binrias so recursivas,
	 * todas necessitam de um parametro de "parada de recurso", ento,
	 * sempre que acharmos um n NULO, retornamos (at porque no d pra
	 * imprimir informaes de um n que no existe.
	 */
	if (!raiz) return ;

	/* essa funo  uma derivao da ordenao 'inorder', onde primeiro
	 * a arvore  percorrida at seu elemento mais esquerdo, e ento  
	 * impressa a informao da raiz, e ento  percorrida suas subrvores
	 * direitas.
	 *
	 * para imprimir de forma 'didtica', essa funo viaja primeiro pelas
	 * folhas direitas, imprime a raiz e ento percorre as subrvores 
	 * esquerdas.
	 */
	imprimir_arvore (raiz->direita, level + 1);

	for (i = 0; i < level * 2; i++ ) putchar(' ');
	printf("%d\n", raiz->info);

	imprimir_arvore (raiz->esquerda, level + 1);
}

/* @ void ordenada (Folha raiz)
 *
 * Transversalizao Ordenada
 * --------------------------
 * Primeiro  visitada a subrvore esquerda, depois a raiz e por ltimo a
 * subrvore direita.
 */
void ordenada (Folha raiz)
{
	if (!raiz) return ;
	
	ordenada (raiz->esquerda);
	printf ("%d ", raiz->info);
	ordenada (raiz->direita);
}

/* @ void preordenada (Folha raiz)
 * 
 * Transversalizao Pr-Ordenada
 * ------------------------------
 * Primeiro  visitada a raiz, depois a subrvore esquerda e por ltimo a 
 * subrvore direita.
 */
void preordenada (Folha raiz)
{
	if (!raiz) return ;
	
	printf ("%d ", raiz->info);
	preordenada (raiz->esquerda);
	preordenada (raiz->direita);
}

/* @ void posordenada (Folha raiz)
 *
 * Transversalizao Ps-Ordenada
 * ------------------------------
 * Primeiro  visitada a subrvore esquerda, depois a subrvore direita e por
 * ltimo a raiz.
 */
void posordenada (Folha raiz)
{
	if (!raiz) return;
	
	posordenada (raiz->esquerda);
	posordenada (raiz->direita);
	printf ("%d ", raiz->info);
}

/******************************************************************************
 * Aqui termina todo o cdigo relacionado s rvores binrias e comea o cdigo
 * relative  interface.
 * 
 */
void cmd (void)
{
	char cmd[CMDSTR], *arg;
	Folha raiz = NULL;
	
	help();
	
	while (1) {
		printf ("ArvoreBinaria> ");
		fgets (cmd, CMDSTR, stdin);
		
		if (!cmd) continue;
	
		switch (*cmd) {
		case 'i':
			arg = &cmd[2];
			arg = strtok(arg, " ");
			while (arg) {
				if (!raiz) raiz = inserir (raiz, atoi(arg));
				else inserir (raiz, atoi(arg));
				
				arg = strtok (NULL, " ");
			}
			break;
		case 'd': 
			arg = &cmd[2];
			arg = strtok(arg, " ");
			while (arg) {
				raiz = deletar (raiz, atoi(arg));
				
				arg = strtok (NULL, " ");
			}
			break;
		case 'm':
			imprimir_arvore (raiz, 0);
			break;
		case 'o':
			ordenada(raiz);
			puts("");
			break;
		case 'r':
			preordenada(raiz);
			puts("");
			break;
		case 'p':
			posordenada(raiz);
			puts("");
			break;
		case 's':
			exit(0);
		case 'h':
			help();
			break;
		default:
			printf("Comando nao reconhecido.\n");
			break;
		}
		
		memset (cmd, 0x0, CMDSTR);
	}
}

/* @ void help (void)
 *
*	i 20	vai inserir o elemento 20 na rvore
 * 	d 20	vai remover o elemento 20 da rvore
 *	m	vai mostrar a rvore na tela
 *	o	transversalizao ordenada
 *	r	transversalizao pr-ordenada
 *	p	transversalizao ps-ordenada
 *	s	sai do programa
 * 	h	mostra a ajuda
 */
void help (void)
{
	printf ("    |    Arvores Binarias\n");
	printf ("    |    Implementacao em C para o Viva O Linux\n");
	printf ("    |    \n");
	printf ("    |    Autor: Enzo Ferber\n");
	printf ("    |    2015\n");
	printf ("    |\n\n");
	
	printf ("\t\tLista de comandos\n");
	printf ("\t\t-----------------\n");
	printf ("\t\ti %%d - Inserir um elemento\n");
	printf ("\t\td %%d - Deletar um elemento\n");
	printf ("\t\tm    - Mostrar a arvore lateralmente\n");
	printf ("\t\to    - Transversalizacao Ordenada\n");
	printf ("\t\tr    - Transversalizacao Pre-Ordenada\n");
	printf ("\t\tp    - Transversalizacao Pos-Ordenada\n");
	printf ("\t\ts    - Sair do programa\n");
	printf ("\t\th    - Mostra a ajuda\n\n");
}
	
int main (void)
{
	cmd ();
	return 0;
}

